/*
 *  drivers/mtd/nand/ts7350.c
 *
 *  Copyright (C) 2007-2008 Technologic Systems (support@embeddedARM.com)
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/arch/hardware.h>
#include <asm/sizes.h>
#include <asm/cacheflush.h>
#include <asm/semaphore.h>
#include <linux/dma-mapping.h>

extern int dma_op (unsigned long dev_addr, unsigned long v_addr, unsigned int len, short rw);

struct partition_info { // little-endian format
  unsigned int bootable:8; // 0x80 = active, 0x00 = don't use for booting
  unsigned int starting_head:8;
  unsigned int starting_csect:8;
  unsigned int starting_xcyl:8;
  // MSB bits 0-5 are sector
  unsigned int system_id:8; // format of partition
  unsigned int ending_head:8;
  unsigned int ending_csect:8;
  unsigned int ending_xcyl:8;
  unsigned int relative_sector:32; // sector offset from start of disk to start of volume
  unsigned int total_sectors:32;
} __attribute__((packed));

struct partition_table {
  struct partition_info part[4];
} __attribute__((__packed__));

struct MBR {
  char bootcode[446];
  struct partition_table pt;
  unsigned int sig55:8;
  unsigned int sigAA:8;
} __attribute__((__packed__));

/*
 * MTD structure for TS7350 board
 */
static struct mtd_info *ts7350_mtd = NULL;


/*
 * Module stuff
 */
static volatile unsigned int ts7350_nandctrl = 0;
static volatile unsigned int ts7350_nanddat = 0;
static spinlock_t lock;
static volatile int busy;

#ifdef CONFIG_MTD_PARTITIONS
static struct mtd_partition partition_mbr[] = {
	{ .name= "whole chip",
		  .offset= 0,
		  .size= 0x20000000 },
	{ .name= "kernel",
		  .offset= 0x20000,
		  .size= 0x400000 },
	{ .name= "initrd",
		  .offset= 0x420000,
		  .size= 0x400000 },
	{ .name= "rootfs",
		  .offset= 0x820000,
		  .size= 0x20000000 - 0x820000 - 0x60000 },
	{ .name= "part4",
		  .offset= 0x20000000 - 0x60000,
		  .size= 0x100000 },
};
#endif

/*
 *	hardware specific access to control-lines
 *
 *	ctrl:
 *	NAND_NCE: bit 0 -> bit 2
 *	NAND_CLE: bit 1 -> bit 1
 *	NAND_ALE: bit 2 -> bit 0
 */
static void ts7350_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
	static unsigned char bits;
	int stall = 0;

	//EWD: upper layers ensure this routine is called in a thread safe
	//manner(e.g. a lock is held to ensure only one thread calls this
	//routine at a time)
	struct nand_chip *chip = mtd->priv;
	/*
	if ((ctrl&NAND_NCE) && !busy) {
		spin_lock_irq(&lock);
		busy=1;
	}
	*/
	if (ctrl & NAND_CTRL_CHANGE) {
		unsigned long addr = ts7350_nandctrl;

		//check if power_ok on rising edge of cs
		if((board_is_prosoft() && (bits & NAND_CLE) == 0) && (ctrl & NAND_CLE)) 
			if(!prosoft_powerok()) stall = 1;

		bits = (ctrl & NAND_NCE) << 2;

		bits |= ctrl & NAND_CLE;
		bits |= (ctrl & NAND_ALE) >> 2;

		__raw_writeb((__raw_readb(addr) & ~0x7) | bits, addr);
	}

	if(stall) {
		printk("Detected power fail, write protecting nand flash\n");
		while(!prosoft_powerok());
	}

	if (cmd != NAND_CMD_NONE) 
		__raw_writeb(cmd, chip->IO_ADDR_W);

	/*
	if (!(ctrl&NAND_NCE) && busy) {
		busy=0;
		spin_unlock_irq(&lock);

		if (need_resched()) schedule();

	}
	*/
}

static void ts73x0_enable_hwecc(struct mtd_info *mtd,int dir) {
	int x;

	/* Enable and empty FIFO */
	x = readl(ts7350_nandctrl);
	while ((x & 0x18) != 0x18) {
		writel(x | 0x8, ts7350_nandctrl);
		x = readl(ts7350_nandctrl);
	}
}

static int ts73x0_calculate_ecc(struct mtd_info *mtd, const u_char *dat, u_char *ecc_code) {
	unsigned dummy,i;


	for (i = 0; i < 8; i++) {
		dummy = readl(ts7350_nandctrl); // *buf;
		dummy =  0xFFFFFF & ((dummy >> 6) | 3);
		*ecc_code++ = (dummy >> 16) & 0xFF;
		*ecc_code++ = (dummy >> 8) & 0xFF;
		*ecc_code++ = (dummy >> 0) & 0xFF;

	}
	return dummy;
}

extern int nand_correct_data(struct mtd_info *mtd, u_char *dat, u_char *read_ecc, u_char *calc_ecc);
int ts73x0_correct_data(struct mtd_info *mtd, u_char *dat,
		                      u_char *read_ecc, u_char *calc_ecc)
{
	int ret = 0, i;
	for (i = 0; i < 8; i++) {
		ret |= nand_correct_data(mtd, dat + (256 * i),
		&read_ecc[i * 3], &calc_ecc[i * 3]);
	}

	return ret;
}
/*
 *	read device ready pin
 */
static int ts7350_device_ready(struct mtd_info *mtd)
{
	return __raw_readb(ts7350_nandctrl) & 0x20;
}

static int nand_wait(struct mtd_info *mtd, struct nand_chip *chip)
{
	unsigned long timeo = jiffies;
	int status, state = chip->state;
	unsigned long addr = ts7350_nandctrl;

	if ((state == FL_ERASING) && (chip->options & NAND_IS_AND))
		chip->cmdfunc(mtd, NAND_CMD_STATUS_MULTI, -1, -1);
	else
		chip->cmdfunc(mtd, NAND_CMD_STATUS, -1, -1);

	//XXX: if someone disabled intterupts for more than 10ms
	//two timer ticks will result back to back with less than 10ms between
	//ticks. Currently this is happening resulting in a premature timeout
	//while waiting for a Nand flash operation to complete. As a result we
	//have removed the timeout and poll forever.
	while (1) {
		if (chip->dev_ready) {
			if (chip->dev_ready(mtd))
				break;
		} else {
			if (chip->read_byte(mtd) & NAND_STATUS_READY)
				break;
		}

		if (need_resched())  {
			if (busy) {
				__raw_writeb((__raw_readb(addr) & ~(NAND_NCE<<2)), addr);
				spin_unlock_irq(&lock);
			}
			schedule();
			if (busy) {
				spin_lock_irq(&lock);
				__raw_writeb((__raw_readb(addr) | (NAND_NCE<<2)), addr);
			}
		}



	}

	status = (int)chip->read_byte(mtd);
	return status;
}

static void ts73x0_read_buf(struct mtd_info *mtd, u_char *buf, int len) {

	while (len > (0xfff * 2)) {
		ts73x0_read_buf(mtd, buf, (0xfff * 2));
		len -= (0xfff * 2);
		buf += (0xfff * 2);
	}

	while(dma_op (0x404, (int)buf, len, MTD_READ) < 0);
}


/*
 * Main initialization routine
 */
static int __init ts7350_init (void)
{
	struct nand_chip *this;
	const char *part_type = 0;
	char *nandaddr;
	char buf[2048];
	int i,got = 0,parts=0;
	unsigned ofs=0;
	struct MBR *mbr = (struct MBR *)buf;

	/* Allocate memory for MTD device structure and private data */
	ts7350_mtd = kmalloc(sizeof(struct mtd_info) + 
			     sizeof(struct nand_chip),
			     GFP_KERNEL);
	if (!ts7350_mtd) {
		printk("Unable to allocate TS7350 NAND MTD device structure.\n");
		return -ENOMEM;
	}

        nandaddr = ioremap(0x600ff000, 0x1000);
	
	/* Get pointer to private data */
	this = (struct nand_chip *) (&ts7350_mtd[1]);
	
	/* Initialize structures */
	memset((char *) ts7350_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));
	
	/* Link the private data with the MTD structure */
	ts7350_mtd->priv = this;
        ts7350_mtd->owner = THIS_MODULE;
	
	/* insert callbacks */
	this->IO_ADDR_R = nandaddr + 0x406;
	this->IO_ADDR_W = nandaddr + 0x404;
	ts7350_nanddat = (unsigned int)nandaddr + 0x404;
	ts7350_nandctrl = (unsigned int)nandaddr + 0x400;

	this->waitfunc = nand_wait;
	this->cmd_ctrl = ts7350_hwcontrol;
	this->dev_ready = ts7350_device_ready;
	this->chip_delay = 15;
	this->options = NAND_USE_FLASH_BBT;
        this->ecc.mode = NAND_ECC_HW;
        this->ecc.size = 2048;
        this->ecc.bytes = 24;
        this->read_buf = ts73x0_read_buf;
	this->ecc.hwctl = ts73x0_enable_hwecc;
	this->ecc.calculate = ts73x0_calculate_ecc;
	this->ecc.correct = ts73x0_correct_data;

	printk("Searching for NAND flash...\n");
	/* Scan to find existence of the device */
	if (nand_scan (ts7350_mtd, 1)) {
		kfree (ts7350_mtd);
		return -ENXIO;
	}
	printk("Scan done!\n");
	for (i=0;i<512;i++) {
	  buf[i] = 0;
	}
	ts7350_mtd->read(ts7350_mtd,0,512,&got,buf);
	printk("read MBR\n");

	if ( mbr->sigAA == 0xAA && mbr->sig55 == 0x55) {
          part_type = "MBR";
	  for (i=0;i<4;i++) {
	    if (mbr->pt.part[i].total_sectors) {
	      parts++;
	      partition_mbr[i+1].offset = 
	        512 * mbr->pt.part[i].relative_sector;
	      if (i < 2) {
	        partition_mbr[i+1].size = 
	          512 * mbr->pt.part[i+1].relative_sector -
	          512 * mbr->pt.part[i].relative_sector;
	      } else {
	        partition_mbr[i+1].size = 
	          512 * mbr->pt.part[i].total_sectors;
	      }
	      ofs = partition_mbr[i+1].offset 
	        + partition_mbr[i+1].size;
	    }
	  }
	} else {
          part_type = "default";
	  parts = 3;
	}
	
	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(ts7350_mtd, partition_mbr, parts + 1);

	spin_lock_init(&lock);
	busy=0;

	/* Return happy */
	return 0;
}
module_init(ts7350_init);

/*
 * Clean up routine
 */
static void __exit ts7350_cleanup (void)
{
	/* Unregister the device */
	del_mtd_device (ts7350_mtd);

	iounmap((void *)ts7350_nandctrl);
	
	/* Free the MTD device structure */
	kfree (ts7350_mtd);
}
module_exit(ts7350_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jesse Off <joff@embeddedARM.com>");
MODULE_DESCRIPTION("MTD driver for Technologic Systems TS-7350 board");
